#define _XOPEN_SOURCE

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "cd_data.h"

/* toto slo mus bt vt ne nejvt jednotliv etzec v jakkoli databzov struktue */
#define TMP_STRING_LEN 125 

/* 
	Nyn vytvome vtov typ typedef s vbry z nabdky. Tento postup je vhodnj, 
	ne definovat jen konstanty jako #define, protoe kompiltor tak me nad 
	promnnmi s uloenmi volbami provdt typov kontroly:
*/
 
typedef enum {
    mo_invalid,
    mo_add_cat,
    mo_add_tracks,
    mo_del_cat,
    mo_find_cat,
    mo_list_cat_tracks,
    mo_del_tracks,
    mo_count_entries,
    mo_exit
} menu_options;

/*
	Dle napeme prototypy loklnch funkc. Nezapomete, e prototypy 
	funkc pro vlastn pstup k databzi jsou uvedeny v souboru cd_data.h:
*/

static int command_mode(int argc, char *argv[]);
static void announce(void);
static menu_options show_menu(const cdc_entry *current_cdc);
static int get_confirm(const char *question);
static int enter_new_cat_entry(cdc_entry *entry_to_update);
static void enter_new_track_entries(const cdc_entry *entry_to_add_to);
static void del_cat_entry(const cdc_entry *entry_to_delete);
static void del_track_entries(const cdc_entry *entry_to_delete);
static cdc_entry find_cat(void);
static void list_tracks(const cdc_entry *entry_to_use);
static void count_all_entries(void);
static void display_cdc(const cdc_entry *cdc_to_show);
static void display_cdt(const cdt_entry *cdt_to_show);
static void strip_return(char *string_to_strip);


/* 
	Konen se dostvme k funkci main. Tu zaneme inicializac struktury 
	current_cdc_entry, ve kter sledujeme aktuln vybranou poloku katalogu CD. 
	Dle provedeme analzu pkazovho dku, oznmme, jak program byl sputn a inicializujeme databzi:
*/

int main(int argc, char *argv[])
{
    menu_options current_option;
    cdc_entry current_cdc_entry;
    int command_result;

    memset(&current_cdc_entry, '\0', sizeof(current_cdc_entry));

    if (argc > 1) {
        command_result = command_mode(argc, argv);
        exit(command_result);
    }

    announce();

    if (!database_initialize(0)) {
        fprintf(stderr, "Je mi lto, nemohu inicializovat databzi\n");
        fprintf(stderr, "Pro vytvoen nov databze pite %s -i\n", argv[0]);
        exit(EXIT_FAILURE);
    }

/* 
	V tomto okamiku jsme pipraveni zpracovat vstup od uivatele. Vstoupme proto do smyky, 
	v n vdy podme o vbr z nabdky a zpracujeme jej a pokraujeme tak dlouho, dokud 
	se uivatel nerozhodne program ukonit. Vimnte si, e pokud je aktuln vybrna njak 
	poloka katalogu, pedvme do funkce show_menu aktuln strukturu current_cdc_entry; 
	dky tomu se rozsah nabdky me zmnit podle toho, jestli je vybrna njak 
	poloka katalogu nebo ne:
*/

    while(current_option != mo_exit) {
        current_option = show_menu(&current_cdc_entry);

        switch(current_option) {
            case mo_add_cat:
                if (enter_new_cat_entry(&current_cdc_entry)) {
                    if (!add_cdc_entry(current_cdc_entry)) {
                        fprintf(stderr, "Nepodailo se pidat novou poloku\n");
                        memset(&current_cdc_entry, '\0',
                                sizeof(current_cdc_entry));
                    }
                }
                break;
            case mo_add_tracks:
                enter_new_track_entries(&current_cdc_entry);
                break;
            case mo_del_cat:
                del_cat_entry(&current_cdc_entry);
                break;
            case mo_find_cat:
                current_cdc_entry = find_cat();
                break;
            case mo_list_cat_tracks:
                list_tracks(&current_cdc_entry);
                break;
            case mo_del_tracks:
                del_track_entries(&current_cdc_entry);
                break;
            case mo_count_entries:
                count_all_entries();
                break;
            case mo_exit:
                break;
            case mo_invalid:
                break;
            default:
                break;
        } /* switch */
    } /* while */

/*
  Po ukonen smyky ve funkci main uzaveme tak databzi a vrtme se zpt do 
	systmovho prosted. vodn pozdrav vypisujeme ve funkci announce:
*/

    database_close();
    exit(EXIT_SUCCESS);
} /* main */

static void announce(void)
{
    printf("\n\nVtejte v ukzkov aplikaci databze \
            s katalogem CD\n");
}

/* 
	Dostvme se k implementaci funkce show_menu. Ta nejprve podle prvnho znaku 
	katalogovho nzvu ov, jestli je vybrna njak poloka katalogu; v takovm 
	ppad nabdne uivateli vbr ze ir mnoiny operac:
	Vimnte si, e v tto verzi na rozdl od pedchozch dvou pklad 
	nevybrme jednotliv volby potenm psmenem, ale slic.
*/ 

static menu_options show_menu(const cdc_entry *cdc_selected)
{
    char tmp_str[TMP_STRING_LEN + 1];
    menu_options option_chosen = mo_invalid;

    while (option_chosen == mo_invalid) {
        if (cdc_selected->catalog[0]) {
            printf("\n\nAktuln poloka: ");
                    printf("%s, %s, %s, %s\n", cdc_selected->catalog,
                    cdc_selected->title,
                    cdc_selected->type,
                    cdc_selected->artist);

            printf("\n");
            printf("1 - pidat nov CD\n");
            printf("2 - hledat CD\n");
            printf("3 - poet CD a stop v databzi\n");
            printf("4 - znovu zadat stopy aktulnho CD\n");
            printf("5 - smazat toto CD a vechny jeho stopy\n");
            printf("6 - vypsat stopy na aktulnm CD\n");
            printf("q - konec\n");
            printf("\nVolba: ");
            fgets(tmp_str, TMP_STRING_LEN, stdin);

            switch(tmp_str[0]) {
                case '1': option_chosen = mo_add_cat; break;
                case '2': option_chosen = mo_find_cat; break;
                case '3': option_chosen = mo_count_entries; break;
                case '4': option_chosen = mo_add_tracks; break;
                case '5': option_chosen = mo_del_cat; break;
                case '6': option_chosen = mo_list_cat_tracks; break;
                case 'q': option_chosen = mo_exit; break;
                }
        }
        else {
            printf("\n\n");
            printf("1 - pidat nov CD\n");
            printf("2 - hledat CD\n");
            printf("3 - poet CD a stop v databzi\n");
            printf("q - konec\n");
            printf("\nVolba: ");
            fgets(tmp_str, TMP_STRING_LEN, stdin);

            switch(tmp_str[0]) {
                case '1': option_chosen = mo_add_cat; break;
                case '2': option_chosen = mo_find_cat; break;
                case '3': option_chosen = mo_count_entries; break;
                case 'q': option_chosen = mo_exit; break;
            }
        }
    } /* while */
    return(option_chosen);
}

/*
	Na nkolika mstech aplikace se budeme uivatele pro kontrolu ptt, jestli skuten 
	chce provst naznaenou operaci. Abychom stejnou otzku nemuseli v kdu zapisovat 
	nkolikrt opakovan, zapeme ji do samostatn funkce get_confirm:
*/

static int get_confirm(const char *question)
{
    char tmp_str[TMP_STRING_LEN + 1];

    printf("%s", question);
    fgets(tmp_str, TMP_STRING_LEN, stdin);
    if (tmp_str[0] == 'Y' || tmp_str[0] == 'y' || tmp_str[0] == 'A' || tmp_str[0] == 'a') {
        return(1);
    }
    return(0);
}

/* 
	Pomoc funkce enter_new_cat_entry me uivatel zadat do katalogu novou poloku. 
	Znak novho dku, naten na konci voln fgets, zde nepotebujeme, take jej ozneme:
	Ve zdrojovm kdu aplikace nepouvme funkci gets, protoe u n nemme jak 
	zabrnit monmu peteen bufferu. Tto funkci se zkrtka vyhbejte!
*/

static int enter_new_cat_entry(cdc_entry *entry_to_update)
{
    cdc_entry new_entry;
    char tmp_str[TMP_STRING_LEN + 1];

    memset(&new_entry, '\0', sizeof(new_entry));

    printf("Zadejte novou poloku do katalogu: ");
    (void)fgets(tmp_str, TMP_STRING_LEN, stdin);
    strip_return(tmp_str);
    strncpy(new_entry.catalog, tmp_str, CAT_CAT_LEN - 1);

    printf("Zadejte titul: ");
    (void)fgets(tmp_str, TMP_STRING_LEN, stdin);
    strip_return(tmp_str);
    strncpy(new_entry.title, tmp_str, CAT_TITLE_LEN - 1);

    printf("Zadejte typ: ");
    (void)fgets(tmp_str, TMP_STRING_LEN, stdin);
    strip_return(tmp_str);
    strncpy(new_entry.type, tmp_str, CAT_TYPE_LEN - 1);

    printf("Zadejte umlce: ");
    (void)fgets(tmp_str, TMP_STRING_LEN, stdin);
    strip_return(tmp_str);
    strncpy(new_entry.artist, tmp_str, CAT_ARTIST_LEN - 1);

    printf("\nNov poloka do katalogu je :-\n");
    display_cdc(&new_entry);
    if (get_confirm("Pidat tuto poloku? ")) {
        memcpy(entry_to_update, &new_entry, sizeof(new_entry));
        return(1);
    }
    return(0);
}

/* Nyn pichz funkce, v n zadvme informace o stopch, a sice enter_new_track_entries. 
	Ta je o nco sloitj ne funkce pro zadvn poloky katalogu, protoe existujc 
	poloka stopy v n me tak zstat nezmnn:
*/

static void enter_new_track_entries(const cdc_entry *entry_to_add_to)
{
    cdt_entry new_track, existing_track;
    char tmp_str[TMP_STRING_LEN + 1];
    int track_no = 1;

    if (entry_to_add_to->catalog[0] == '\0') return;

    printf("\nAktualizuji stopy na CD %s\n", entry_to_add_to->catalog);
    printf("Stiskem Enter potvrdte nezmnn stvajc popis,\n");
    printf(" psmenem d smaete tuto a vechny zbvajc stopy,\n");
    printf(" nebo zadejte nov popis stopy\n");

    while(1) {

/*
	Nejprve musme zkontrolovat, jestli existuje stopa s aktulnm slem stopy. Podle vsledk zmnme text vzvy:
*/

        memset(&new_track, '\0', sizeof(new_track));
        existing_track = get_cdt_entry(entry_to_add_to->catalog,
                                       track_no);
        if (existing_track.catalog[0]) {
            printf("\tStopa %d: %s\n", track_no,
                    existing_track.track_txt);
            printf("\tNov text: ");
        }
        else {
            printf("\tPopis stopy %d: ", track_no);
        }
        fgets(tmp_str, TMP_STRING_LEN, stdin);
        strip_return(tmp_str);

/* 
	Jestlie ke stop neexistoval dn zznam a uivatel nevytvoil ani nov zznam, 
	pedpokldme, e uivatel poaduje ukonen zadvn stop:
*/
        if (strlen(tmp_str) == 0) {
            if (existing_track.catalog[0] == '\0') {
                /* nen existujc poloka, take konme zadvn */
                break;
            }
            else {
                /* peskoit stvajc poloku a pejt na dal stopu */
                track_no++;
                continue;
            }
        }

/* 
	Zad-li uivatel samotn psmeno d, odstranme tm aktuln stopu a vechny 
	nsledujc stopy. Pokud poadovanou stopu nelze najt, vrt 
	funkce del_cdt_entry hodnotu false:
*/

        if ((strlen(tmp_str) == 1) && tmp_str[0] == 'd') {
                /* odstranit tuto a zbvajc stopy */
            while (del_cdt_entry(entry_to_add_to->catalog, track_no)) {
                track_no++;
            }
            break;
        }

/* 
	Jako dal ns ek sek kdu pro pidn nov stopy nebo pro aktualizaci stvajc
	stopy. Zde vytvome novou strukturu new_track typu cdt_entry a pot zavolme 
	databzovou funkci add_cdt_entry, kter ji pid do databze:
*/

        strncpy(new_track.track_txt, tmp_str, TRACK_TTEXT_LEN - 1);
        strcpy(new_track.catalog, entry_to_add_to->catalog);
        new_track.track_no = track_no;
        if (!add_cdt_entry(new_track)) {
            fprintf(stderr, "Pidn nov stopy se nepodailo\n");
            break;
        }
        track_no++;
    } /* while */
}

/* 
	Funkce del_cat_entry slou k odstrann poloky katalogu. Nesmme pitom 
	dopustit, aby v databzi existovaly stopy bez pslun poloky katalogu:
*/

static void del_cat_entry(const cdc_entry *entry_to_delete)
{
    int track_no = 1;
    int delete_ok;

    display_cdc(entry_to_delete);
    if (get_confirm("Odstranit tuto poloku a vechny jej stopy? ")) {
        do {
            delete_ok = del_cdt_entry(entry_to_delete->catalog,
                                      track_no);
            track_no++;
        } while(delete_ok);

        if (!del_cdc_entry(entry_to_delete->catalog)) {
            fprintf(stderr, "Nepodailo se odstranit poloku\n");
        }
    }
}

/* Dal funkce je pomocn a jejm kolem je odstranit veker stopy dan poloky katalogu: */
static void del_track_entries(const cdc_entry *entry_to_delete)
{
    int track_no = 1;
    int delete_ok;

    display_cdc(entry_to_delete);
    if (get_confirm("Odstranit stopy pro tuto poloku? ")) {
        do {
            delete_ok = del_cdt_entry(entry_to_delete->catalog, track_no);
            track_no++;
        } while(delete_ok);
    }
}

/* 
	V aplikaci napeme tak velice jednoduchou funkci pro vyhledvn v katalogu. 
	Nechme uivatele zadat etzec, a pot z katalogu vyhledme poloky, kter 
	jej obsahuj. Vslednch poloek me bt vce, a proto je uivateli 
	vypeme v nabdce:
*/

static cdc_entry find_cat(void)
{
    cdc_entry item_found;
    char tmp_str[TMP_STRING_LEN + 1];
    int first_call = 1;
    int any_entry_found = 0;
    int string_ok;
    int entry_selected = 0;

    do {
        string_ok = 1;
        printf("Zadejte etzec, hledan mezi polokami katalogu: ");
        fgets(tmp_str, TMP_STRING_LEN, stdin);
        strip_return(tmp_str);
        if (strlen(tmp_str) > CAT_CAT_LEN) {
            fprintf(stderr, "Je mi lto, etzec je pli dlouh, maximum %d \
                    znak\n", CAT_CAT_LEN);
            string_ok = 0;
        }
    } while (!string_ok);

    while (!entry_selected) {
        item_found = search_cdc_entry(tmp_str, &first_call);
        if (item_found.catalog[0] != '\0') {
            any_entry_found = 1;
            printf("\n");
            display_cdc(&item_found);
            if (get_confirm("Tato poloka? ")) {
                entry_selected = 1;
            }
        }
        else {
            if (any_entry_found) printf("Je mi lto, dal poloky nebyly nalezeny\n");
            else printf("Je mi lto, nic nenalezeno\n");
            break;
        }
    }
    return(item_found);
}

/* Pomocn funkce list_tracks vype vechny stopy k dan poloce katalogu: */
static void list_tracks(const cdc_entry *entry_to_use)
{
    int track_no = 1;
    cdt_entry entry_found;

    display_cdc(entry_to_use);
    printf("\nStopy\n");
    do {
        entry_found = get_cdt_entry(entry_to_use->catalog,
                                    track_no);
        if (entry_found.catalog[0]) {
            display_cdt(&entry_found);
            track_no++;
        }
    } while(entry_found.catalog[0]);

    (void)get_confirm("Stisknte Enter");
} /* list_tracks */

/* Dal funkce, count_all_entries, spot vechny stopy v databzi: */
static void count_all_entries(void)
{
    int cd_entries_found = 0;
    int track_entries_found = 0;
    cdc_entry cdc_found;
    cdt_entry cdt_found;
    int track_no = 1;
    int first_time = 1;
    char *search_string = "";

    do {
        cdc_found = search_cdc_entry(search_string, &first_time);
        if (cdc_found.catalog[0]) {
            cd_entries_found++;
            track_no = 1;
            do {
                cdt_found = get_cdt_entry(cdc_found.catalog, track_no);
                if (cdt_found.catalog[0]) {
                    track_entries_found++;
                    track_no++;
                }
            } while (cdt_found.catalog[0]);
        }
    } while (cdc_found.catalog[0]);

    printf("Nalezeno %d CD, celkem %d stop\n", cd_entries_found,
                track_entries_found);

    (void)get_confirm("Stisknte Enter");
}

/* Dostvme se k funkci display_cdc, jejm kolem je zobrazen poloky katalogu: */
static void display_cdc(const cdc_entry *cdc_to_show)
{
    printf("Katalog: %s\n", cdc_to_show->catalog);
    printf("\ttitul: %s\n", cdc_to_show->title);
    printf("\ttyp: %s\n", cdc_to_show->type);
    printf("\tumlec: %s\n", cdc_to_show->artist);
}

/* a k funkci display_cdt, kter zobrazuje poloku jedn stopy */
static void display_cdt(const cdt_entry *cdt_to_show)
{
    printf("%d: %s\n", cdt_to_show->track_no, cdt_to_show->track_txt);
}

/*
	Nsleduje pomocn funkce strip_return, kter z etzce odstran koncov znak 
	novho dku. Nezapomete, e Linux stejn jako Unix oznauje konec 
	dku jedinm znakem LF:
*/

static void strip_return(char *string_to_strip)
{
    int len;

    len = strlen(string_to_strip);
    if (string_to_strip[len - 1] == '\n') string_to_strip[len - 1] = '\0';
}

/* 
	Nzev command_mode skrv funkci pro analzu argument pkazovho dku. Voln 
	funkce getopt zde zajiuje, e n program bude pebrat argumenty podle 
	obvyklch konvenc systmu Linux:
*/

static int command_mode(int argc, char *argv[])
{
    int c;
    int result = EXIT_SUCCESS;
    char *prog_name = argv[0];

    /* tyto extern promnn pouv funkce getopt */
    extern char *optarg;
    extern int optind, opterr, optopt;

    while ((c = getopt(argc, argv, ":i")) != -1) {
        switch(c) {
            case 'i':
                if (!database_initialize(1)) {
                    result = EXIT_FAILURE;
                    fprintf(stderr, "Nepodailo se inicializovat databzi\n");
                }
                break;
            case ':':
            case '?':
                default:
                fprintf(stderr, "Voln: %s [-i]\n", prog_name);
                result = EXIT_FAILURE;
                break;
        } /* switch */
    } /* while */
    return(result);
}

